/*
 * ext/ctype/ctype.c
 * 
 * by Eivind Eklund <eivind@FreeBSD.org>
 */

#include <ctype.h>

#include "ruby.h"
#include "rubyio.h"

static VALUE mFmt;

/*-------------------------- module Ctype --------------------------*/

#define E_EOL	1		/* End of line */
#define E_EOS	2		/* End of string */
#define E_CTRL	4		/* Hit control character */
#define E_SPACE	8		/* Hit space */
#define E_NL	16		/* Hit newline */

#define EVENT(baseoffset, lineoffset, len, width, c) \
	((isspace(c) ? E_SPACE : 0)| \
	 (isctrl(c) ? E_CTRL : 0)| \
	 ((baseoffset) >= (len) ? E_EOL : 0)| \
	 ((lineoffset) >= (width) ? E_EOS : 0)| \
	 ((c) == '\n' ? E_NL : 0))

/*
 * Format a line with word wrapping to a specific length.
 * Will return a array of ruby strings, each of them with length <= <width>.
 * Fails for <width>s <= 0 (throwing an exception).
 * Words that are longer than <width> will be split into several <width> width
 * lines, and more wrods wrapped onto the last of these.
 * Any \n's in the input will be considered a forced line termination, and
 * result in a new line being started (but the \n will NOT be included in the
 * output.)
 * Any control characters (0-31 and 128-159) will be considered whitespace,
 * and a run of these will be replaced with a single space.
 * Lines will be broken at whitespace, and on line termination, all whitespace
 * at the break will be removed.
 */
static VALUE
ruby_wordwrap(
	VALUE	ps,		/* String to split into lines */
	VALUE	iwidth		/* Width of lines to split into */

) {
	/* The internal state machine for this string formatting function is
	 * based on the EVENT macro above classifying what happens when the
	 * next character is added.
	 * Note that the machine optimistically copies all the time, and just
	 * terminate the strings at an earlier point and backtracks when it
	 * finds that it was not able to process more of the string into the
	 * present line.
	 * This is hopefully the fastest approach to the normal formatting
	 * case, as MOST words will fit nicely inside the string, but some may
	 * not.  The only major optimization I can see is to force
	 * pre-fetching of cache lines for the string, but this is highly
	 * dependent on the cache architecture of the target machine, and
	 * would somewhat complicate the code.  If you need extreme formatting
	 * performance, this might be worthwhile.  It may also be worthwhile
	 * to pre-scan parts of the string in order to see how it should be
	 * split.  However, the main part of this kind of string formatting
	 * for Ruby is likely to be the allocation of the lines, anyway.
	 */
	const char *s = RSTRING(ps)->ptr;	/* XXX Single RSTRING cast for efficiency? */
	size_t len = RSTRING(ps)->len;
	char *strp;				/* Output string pointer */
	VALUE array;
	VALUE outstr;
	size_t start_of_word;			/* At what offset do the word
						   we are presently handling
						   start? */
	size_t start_of_line;			/* Same, but for the line we
						   are handling. */
	size_t i;
	size_t width = NUM2INT(iwidth);		/* XXX What is the correct conversion? */
	int e,e_next;				/* The event type we presently hit */
	NEWOBJ(outstr, struct RString);
	OBJSETUP(outstr, rb_cString, T_STRING);

	outstr->ptr = ALLOC_N(char, width+1);
	outstr->len = width;	/* XXX This may not be the ideal way to init? */
	outstr->orig = 0;	/* XXX Should this be copied off rb_str_new,
				   or is it internal and we should use
				   rb_str_new? */
	outstr->ptr[width] = '\0';

	start_of_word = 0;
	strp = outstr->ptr;

	/*
	 * Special case empty strings, in order to remove complexity from the
	 * state machine
	 */
	if (len == 0) {
		outstr->len = 0;
		EE_PUSH(outstr);
		return array;
	}


	/* State machine for handling string formatting */

	/* Start state: First character in a string */
	e = EVENT(i, 0, len, width, s[0]);
	if (e & E_CTRL) {
		/* Control character - collapse more control characters if
		 * available */
		e_next = EVENT(i+1, 1, len, width, s[i+1]);
	}



	for (i = 0; i < len; ) {
		/* Find more words until we have a line, collapsing control
		 * characters to whitespace as we go.  If we hit control
		 * characters, we will replace these with whitespace. */
		size_t wbase = 0;
		size_t j;

		for (j = i; j < len; j++) {
			if (isctrl(s[j])) {
				/*
				 * Control character - should be output as a
				 * space, joined with any other control
				 * characters that follow it, and abandoned if
				 * these occur as the end of the string we are
				 * formatting.
				 */
				int k;
				for (k = j; k < len && isctrl(s[k]); k++)
					;
				if (k == len) {
					/*
					 * We are done - push outstr on the
					 * output array and return the array.
					 */
					XXX
				} else {
					/* Word goes from i through j-1, with
					 * j through k-1 containing control
					 * characters.  If we have space for
					 * an entire word plus a space, put it
					 * on the outstr.  If not, put the
					 * word on the outstr (as there IS
					 * space for that), terminate the
					 * outstr, put it on the array, start
					 * a new outstr, and point everything
					 * at that.  The generated space is
					 * dropped.
					 */
					XXX
				}
			} else if (s[j] == '\n') {
				/*
				 * Newline - we should start a new output
				 * string, and put the one we have on the
				 * output array.
				 */
				/*
				 * Push present string onto array, generate
				 * new string, skip newline.
				 */
			} else if (isspace(s[j])) {
				/* Space character - should be output iff we
				 * get enough space for the next word */
				/* Copy the present word (including previous
				 * whitespace) onto the string, scan all
				 * upcoming whitespace, and if we hit length
				 * limit, start new line.
				 */
			} else {
				/* Normal character */
				/* If we have pushed words onto the present
				 * string (its length is more than 0): */
					/* If we go beyond the length limit, terminate
				 	* present string and push it onto the
					* array.
					* Set start of copying to be from the
					* start of the present word (we DID
					* record that, right?).
					*/
				/* If we have not pushed words onto the
				 * present string (its length is 0): */
					/* If we go beyond the length limit
					 * (from start of word to the last
					 * character scanned), copy the string
					 * up to the limit to the output, and
					 * start a new string.  Reset the
					 * start-of-word pointer to the
					 * present position.
					 */
			}
		}
	}
}

/*------------------------- Initialization -------------------------*/
void
Init_fmt()
{
/* Extra R first to avoid reserved namespace 'STR*' */
#undef RSTRINGIZE
#define RSTRINGIZE(str) #str
    mFmt = rb_define_module("Fmt");
#define reg_is(a) rb_define_module_function(mCtype, RSTRINGIZE(a) "?", ruby_is##a, 1);
reg_is(alnum)
reg_is(alpha)
reg_is(ascii)
reg_is(blank)
reg_is(cntrl)
reg_is(digit)
reg_is(graph)
reg_is(hexnumber)
reg_is(ideogram)
reg_is(lower)
reg_is(number)
reg_is(phonogram)
reg_is(print)
reg_is(punct)
reg_is(rune)
reg_is(space)
reg_is(special)
reg_is(upper)
reg_is(xdigit)
#undef reg_is
#undef RSTRINGIZE
}
