/*
 * ext/ctype/ctype.c
 * 
 * by Eivind Eklund <eivind@FreeBSD.org>
 */

#include <ctype.h>

#include "ruby.h"
#include "rubyio.h"

static VALUE mFmt;

/*-------------------------- module Ctype --------------------------*/

/*
 * Format a line with word wrapping to a specific length.
 * Will return a array of ruby strings, each of them with length <= <width>.
 * Fails for <width>s <= 0 (throwing an exception).
 * Words that are longer than <width> will be split into several <width> width
 * lines, and more wrods wrapped onto the last of these.
 * Any \n's in the input will be considered a forced line termination, and
 * result in a new line being started (but the \n will NOT be included in the
 * output.)
 * Any control characters (0-31 and 128-159) will be considered whitespace,
 * and a run of these will be replaced with a single space.
 * Lines will be broken at whitespace, and on line termination, all whitespace
 * at the break will be removed.
 */
static VALUE
ruby_wordwrap(
	VALUE	ps,		/* String to split into lines */
	VALUE	width		/* Width of lines to split into */

) {
	const char *s = RSTRING(ps)->ptr;	/* XXX Single RSTRING cast for efficiency? */
	size_t len = RSTRING(ps)->len;
	char *strp;				/* Output string pointer */
	VALUE array;
	VALUE outstr;
	size_t start_of_word;			/* At what offset do the word
						   we are presently handling
						   start? */
	size_t start_of_line;			/* Same, but for the line we
						   are handling. */
	size_t i;
	NEWOBJ(outstr, struct RString);
	OBJSETUP(outstr, rb_cString, T_STRING);

	outstr->ptr = ALLOC_N(char, width+1);
	outstr->len = width;	/* XXX This may not be the ideal way to init? */
	outstr->orig = 0;	/* XXX Should this be copied off rb_str_new,
				   or is it internal and we should use
				   rb_str_new? */
	outstr->ptr[width] = '\0';

	start_of_word = 0;
	strp = outstr->ptr;

	for (i = 0; i < len; ) {
		/* Find more words until we have a line, collapsing control
		 * characters to whitespace as we go.  If we hit control
		 * characters, we will replace these with whitespace. */
		size_t wbase = 0;
		size_t j;

		for (j = i; j < len; j++) {
			if (isctrl(s[j])) {
				/*
				 * Control character - should be output as a
				 * space, joined with any other control
				 * characters that follow it, and abandoned if
				 * these occur as the end of the string we are
				 * formatting.
				 */
				int k;
				for (k = j; k < len && isctrl(s[k]); k++)
					;
				if (k == len) {
					/*
					 * We are done - push outstr on the
					 * output array and return the array.
					 */
					XXX
				} else {
					/* Word goes from i through j-1, with
					 * j through k-1 containing control
					 * characters.  If we have space for
					 * an entire word plus a space, put it
					 * on the outstr.  If not, put the
					 * word on the outstr (as there IS
					 * space for that), terminate the
					 * outstr, put it on the array, start
					 * a new outstr, and point everything
					 * at that.  The generated space is
					 * dropped.
					 */
					XXX
				}
			} else if (s[j] == '\n') {
				/*
				 * Newline - we should start a new output
				 * string, and put the one we have on the
				 * output array.
				 */
				/*
				 * Push present string onto array, generate
				 * new string, skip newline.
				 */
			} else if (isspace(s[j])) {
				/* Space character - should be output iff we
				 * get enough space for the next word */
				/* Copy the present word (including previous
				 * whitespace) onto the string, scan all
				 * upcoming whitespace, and if we hit length
				 * limit, start new line.
				 */
			} else {
				/* Normal character */
				/* If we have pushed words onto the present
				 * string (its length is more than 0): */
					/* If we go beyond the length limit, terminate
				 	* present string and push it onto the
					* array.
					* Set start of copying to be from the
					* start of the present word (we DID
					* record that, right?).
					*/
				/* If we have not pushed words onto the
				 * present string (its length is 0): */
					/* If we go beyond the length limit
					 * (from start of word to the last
					 * character scanned), copy the string
					 * up to the limit to the output, and
					 * start a new string.  Reset the
					 * start-of-word pointer to the
					 * present position.
					 */
			}
		}
	}
}

/*------------------------- Initialization -------------------------*/
void
Init_fmt()
{
/* Extra R first to avoid reserved namespace 'STR*' */
#undef RSTRINGIZE
#define RSTRINGIZE(str) #str
    mFmt = rb_define_module("Fmt");
#define reg_is(a) rb_define_module_function(mCtype, RSTRINGIZE(a) "?", ruby_is##a, 1);
reg_is(alnum)
reg_is(alpha)
reg_is(ascii)
reg_is(blank)
reg_is(cntrl)
reg_is(digit)
reg_is(graph)
reg_is(hexnumber)
reg_is(ideogram)
reg_is(lower)
reg_is(number)
reg_is(phonogram)
reg_is(print)
reg_is(punct)
reg_is(rune)
reg_is(space)
reg_is(special)
reg_is(upper)
reg_is(xdigit)
#undef reg_is
#undef RSTRINGIZE
}
